/* Compress 16-bit floating point photocurrents (4-bit exponent, 12-bit mantissa) using Rice compression. */
/* Data are read in blocks from a binary input file containing 2048x2048 16-bit photocurrents.
 * The input file also contains (at the beginning of the file) a block of offsets, one for each data block.
 * The input file is generated by the pack16b_v02 program, see pack16b_v02.c for details.
 * 
 * Usage: comp16b_v02 infile outfile npix_per_block method
 * 
 * Compressed data are written to a binary file with the following format:
 * File Header: 16-bit nblocks, 16-bit npix_per_block
 * Data: nblocks compressed photocurrent blocks, in the following format:
 *       12-bit signed offset. The least-significant 11 bits should be read as a positive value.
 *                             If the ms-bit is 1, the value should be multipied by -1.
 *       4-bit Nbits. This is the number of bits used to represent the small-value samples.
 *       8-bit CFSize. This is the length in bytes of the Cflags block.
 *       Cflags: a block of compressed flags, CFSize bytes in length.
 *       Cdata: a block of compressed photocurrent data. The size of the Cdata block
 *                              can be calculated from npix_per_block, Nbits, and the
 *			        decompressed flags.
 * See the SPHEREx SUR Algorithm and Data Formats document for further details.
 *
 * The method command-line argument can be 0 or 1. This chooses from one of two methods for implementing
 * the Rice compression. The compression results are almost identical, but method 1 uses fewer CPU cycles.
 *
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>

#define IMSIZE 2048 * 2048
		     
int main(argc,argv)
    int argc;
    char *argv[];
{
    int infile, outfile;    /* file descriptors */
    int *minimums;  /* buffer containing a block of minimums, one for each block of 16-bit photocurrents */
    unsigned short *inbuf;  /* buffer containing a block of 16-bit photocurrents	*/
    unsigned char *cbuf;    /* buffer containing compressed block */
    unsigned short min;
    int nbits_method;
    int npix;		    /* num of pixels in each block to be compressed */
    int nblocks;	    /* num of blocks per image */
    int cbytes;		    /* number of bytes in compressed block */
    int ii,jj;
    short tmp;

    int write_block(int file, unsigned short *buf, int nbytes);
    int compress_block(unsigned short *ibuf, unsigned char *cbuf, int npix, int min, int nbits_method);
    int write_compressed_block(int file, unsigned char *buf, int nbytes);
    void reset_cbuf();

    if (argc!=5) {
       fprintf(stderr,"Usage: %s infile outfile npix_per_block method\n",argv[0]);
       exit(1);
    }

    if ((infile=open(argv[1],O_RDONLY)) == -1) {
       fprintf(stderr,"Error: %s could not open infile\n",argv[0]);
       exit(1);
    }

    if ((outfile=open(argv[2],O_CREAT | O_RDWR | O_TRUNC,S_IRUSR | S_IWUSR)) == -1) {
       fprintf(stderr,"Error: %s could not open outfile\n",argv[0]);
       exit(1);
    }

    npix = atoi(argv[3]);
    if (npix<0) {
	fprintf(stderr,"Error: %s npix_per_block %d is negative\n",argv[0],npix);
	exit(1);
    }
    if (npix>IMSIZE) {
	fprintf(stderr,"Error: %s npix_per_block %d is greater than image size %d\n",argv[0],npix,IMSIZE);
	exit(1);
    }
    if (npix>65535) {
	fprintf(stderr,"Error: %s npix_per_block %d is greater than 65535\n",argv[0],npix);
	exit(1);
    }
    if (IMSIZE % npix) {
	fprintf(stderr,"Error: %s image size %d must be divisible by npix_per_block %d\n",argv[0],IMSIZE,npix);
	exit(1);
    }

    nbits_method = atoi(argv[4]);
    if ((nbits_method !=0)&&(nbits_method !=1)) {
	fprintf(stderr,"Error: %s method is %d: It must be 0 or 1\n",argv[0],nbits_method);
	exit(1);
    }

    nblocks = IMSIZE / npix;
/*    fprintf(stderr,"nblocks= %d\n",IMSIZE / npix); */

    minimums= (int *) malloc(nblocks * sizeof(int));
    inbuf= (unsigned short *) malloc(npix * sizeof(unsigned short));
    cbuf= malloc(10 * npix * sizeof(unsigned short));		/* make it 10x bigger, in case compressed size is larger */

    /* write file header - npix and nblocks */

    tmp = nblocks;
    if (write(outfile,&tmp,sizeof(short)) != sizeof(short)) {
	fprintf(stderr,"%s: Error: could not write to compressed file.\n",argv[0]);
	exit(-1);
    }
    tmp = npix;
    if (write(outfile,&tmp,sizeof(short)) != sizeof(short)) {
	fprintf(stderr,"%s: Error: could not write to compressed file.\n",argv[0]);
	exit(-1);
    }

    if (read(infile,minimums,nblocks * sizeof(int)) != (nblocks * sizeof(int))) {
	fprintf(stderr,"Error: %s read minimums error %d\n",argv[0],ii);
	exit(1);
    }

    for (ii=0;ii<nblocks;ii++) {
	if (read(infile,inbuf,npix * sizeof(unsigned short)) != (npix * sizeof(unsigned short))) {
	    fprintf(stderr,"Error: %s read block error %d\n",argv[0],ii);
	    exit(1);
	}
/*	if (write_block(outfile,inbuf,npix * sizeof(unsigned short)) != 0) {exit(1);}    */

/*
	for (jj=0;jj<8;jj++) {
	    fprintf(stdout,"%04x\n",inbuf[jj]);
	}
*/

	cbytes = compress_block(inbuf,cbuf,npix,minimums[ii],nbits_method);

	if (write_compressed_block(outfile,cbuf,cbytes) != 0) {exit(1);} 

	reset_cbuf();

    }

    close(infile);
    close(outfile);
}

int write_block(int ofile,unsigned short *buf,int nbytes) {
    int errsv,nwrite;

    nwrite = write(ofile,buf,nbytes);
    errsv = errno;
    if (nwrite != nbytes) {
	fprintf(stderr,"write_block: Error: could not write to binfile. errno=%d\n",errsv);
	return(-1);
    }
    return(0);
}

int bitsInBuffer = 0;
int bytesWritten = 0;
unsigned char bits = 0;

void reset_cbuf() {
    bitsInBuffer = 0;
    bytesWritten = 0;
    char bits = 0;
}

int compress_block(unsigned short *inbuf,unsigned char *cbuf,int npix, int min, int nbits_method) {
    int ii;
    int nbits;			/* number of bits N used to represent the small-value samples */
    int best_nbits0;		/* best value of N for this block,method0 */
    int av_nbits1;		/* av value of N for this block,method1 */
    int best_nbits1;		/* best value of N for this block,method1 */
    int nfit,nfitx,nfity;	/* number of samples that fit into N bits */
    int nsize,nsizex,nsizey;	/* size of compressed samples for the current value of nbits */
    int best_nsize;		/* smallest (best) value for nsize for this block */
    unsigned short flags[npix];	/* 1-bit flag for each sample, indicating whether the sample fits into N bits */
    int cfsize = 0;		/* size in bytes of the compressed flag data in cfbuf */
    int maxval[11] = {0,1,3,7,15,31,63,127,255,511,1023};	/* max val that can be represented in N bits */

    void WriteBits(unsigned short inputBits, int numBits, unsigned char *output);
    void WriteFlush(unsigned char *obuf);
    int compress_flags(unsigned short *flags,unsigned char *output,int npix);

    best_nbits0 = 4;			/* default value */
    best_nsize = npix * 16;


    int nsizes[10];

    /* This is method 0. We currently do it even if method 1 is selected, so we can compare the results with method 1 */
    for (nbits=2;nbits<11;nbits++) {	/* loop thru data for each value of nbits */
	    nfit = 0;
	    for (ii=0;ii<npix;ii++) {	/* figure out which samples can fit into N bits */
		if (inbuf[ii]<=maxval[nbits])
		    nfit++;
	    }
	    nsize = (nfit * nbits) + ((npix-nfit) * 16);	/* compressed data size for current N */
	    nsizes[nbits] = nsize;
	    if (nsize < best_nsize) {
		best_nbits0 = nbits;
		best_nsize = nsize;
	    }
    }

    int av=0; int n0=0; int n1=0; int n2=0; int dec = 0; int inc = 0;


    /* This is method 1. We use log2 of the average value of the <1024 photocurrents to select the best nbits */
    /* method 1 uses two loops through the pixel data, whereas method 0 uses 9 */
    for (ii=0;ii<npix;ii++) {		/* first loop thru the data */
	if (inbuf[ii] < 1024) {
	    av+=inbuf[ii];
	    n0++;
	}
    }
    if (n0>0) {
       av = av/n0;	    /* integer divide */
       if (av==0) {
          av = 1;
	  av_nbits1=2;
       } else {
          av_nbits1 = (int)(log2(av)) +1;	/* In the FSW this can be done with a lookup table */
       }
    } else {
       /* fprintf(stderr,"Warning: compress_block: no samples < 1024\n");      */
       /* all the samples in the block are >=1024, so there will be no compression...they will all be written as 16-bit values */
       av = 1;
       av_nbits1 = 2;
    }

    if (av_nbits1 < 2) av_nbits1 = 2;	    /* check for out-of-bounds conditions */
    if (av_nbits1 > 10) av_nbits1 = 10;

    /* Unfortunately, av_nbits1 does not always give the best answer. It is sometimes off by 1.
     * So we have to do one more loop thru the data to check for this */
    if (av_nbits1 < 10) {
	    nfitx = 0;
	    nfity = 0;
	    for (ii=0;ii<npix;ii++) {			/* second loop thru the data */
		if (inbuf[ii]<=maxval[av_nbits1])
		    nfitx++;
		if (inbuf[ii]<=maxval[av_nbits1+1])
		    nfity++;
	    }
	    nsizex = (nfitx * av_nbits1) + ((npix-nfitx) * 16);	
	    nsizey = (nfity * (av_nbits1+1)) + ((npix-nfity) * 16);	
	    if (nsizey < nsizex) {
		best_nbits1 = av_nbits1+1;
	    } else {
		best_nbits1 = av_nbits1;
	    }
    } else {
        best_nbits1 = av_nbits1;
    }
    if (best_nbits1 < 2) best_nbits1 = 2;
    if (best_nbits1 > 10) best_nbits1 = 10;


    /* set nbits based on nbits_method */
    if (nbits_method==0) {
	nbits = best_nbits0;
    } else {
	nbits = best_nbits1;
    }

    fprintf(stdout,"%2d %2d %2d      %4d % 3d %5.1f %4d %4d %4d %4d %4d %4d %4d   %5d\n",
	    nbits,best_nbits0,best_nbits1,  av, n0, log2(av) +1,
	    nsizes[2],nsizes[3],nsizes[4],nsizes[5],nsizes[6],nsizes[7],nsizes[8], min);


    /* now actually do the compression, using the value for nbits that we have selected */

    for (ii=0;ii<npix;ii++) {   /* figure out which samples can fit into nbits bits */
        if (inbuf[ii]<=maxval[nbits])
            flags[ii]=0;
	else {
            flags[ii]=1;    /* use 1 for the larger samples, since there should be fewer of these */
	}
    }

    /* write block header to cbuf */
    /* The minimum sample - Needs to fit into 12 bits. It should have been bounded within -2047 to +2047 inclusive earlier. */
    if ((min<-2047) || (min>2047)) {
	fprintf(stderr,"Warning: compress_block: minimum is out of bounds: %d\n",min);
    }
    if (min < 0) {
       min = -1 * min;
       cbuf[0] = (char)( ((min>>4) & 0x7F) + 0x80 );	/* the MS-bit of the 12 bits is the sign bit */
    } else {
       cbuf[0] = (char)( (min>>4) & 0xFF );
    }
    cbuf[1] = (char)( (min & 0xF) <<4 );		/* the 4 LS-bits of the min */

    /* The number of bits used to represent the small-value samples */
    cbuf[1] = (char)(cbuf[1] + (nbits & 0xF));

    cbuf[2] = 0;	/* Reserve space in the buffer for the size of compressed flags data */

    cbuf[3] = 0;	/* Reserve space in the buffer for other meta data (e.g. size of compressed photocurrents data */
    cbuf[4] = 0;
    cbuf[5] = 0;

    bytesWritten = 6;

    /* compress and write the flags bitfield */
    /* need to skip this if nbits_method is non-zero, and put something in the header to signal this */
    cfsize = compress_flags(flags,cbuf,npix);
    if (cfsize>0xFF) {
	fprintf(stderr,"Warning: compress_block: cfsize is too large: %d\n",cfsize);
	cfsize = 0xFF;	/* should never happen. */
    }

    /* fill-in proper value for cfsize in cbuf */
    cbuf[2] = (char)(cfsize & 0xFF); 

    /* Write compressed photocurrents data to cbuf. */
    /* The npix data items, using nbits bits for the small-value items, 16 bits for the rest */    
    for (ii=0;ii<npix;ii++) {	
	if (flags[ii]==0) {
	    WriteBits(inbuf[ii],nbits,cbuf);
	} else {
	    WriteBits(inbuf[ii],16,cbuf);
	}
    }

    WriteFlush(cbuf);

/*
    for (ii=0;ii<npix;ii++) {
	WriteBits(inbuf[ii],16,cbuf);
    }
    WriteFlush(cbuf);
*/
    
    return(bytesWritten);
}

int compress_flags(unsigned short *flags,unsigned char *obuf,int npix) {
    int ii,jj;
    int nbits=0;
    int nbytes;
    unsigned short threebits;
/*    unsigned short codeword[8] = {0,4,5,28,6,29,30,31};	/* 3-bit encoding codewords from Rice paper, Table 2-1 */
    unsigned short codeword[8] = {0,1,5,7,3,23,15,31};	/* write them with the bits reversed, so the decoding works easier */
    int codebits[8] = {1,3,3,5,3,5,5,5};	/* number of bits for each codeword, including zero-padding */

    void WriteBits(unsigned short inputBits, int numBits, unsigned char *output);
    void WriteFlush(unsigned char *obuf);

    jj= 0 ;
    threebits = 0;
    for (ii=0;ii<npix;ii++) {	/* the npix flags, one bit per flag */
	threebits |= flags[ii] << jj;
	jj++;
	if (jj==3) {
	    WriteBits(codeword[threebits],codebits[threebits],obuf);
	    nbits+=codebits[threebits];
	    threebits=0;
	    jj=0;
	}
    }
    if (jj!=0) {
	WriteBits(codeword[threebits],codebits[threebits],obuf);
	nbits+=codebits[threebits];
    }
    WriteFlush(obuf);

    nbytes = nbits/8;
    if (nbits % 8 !=0 ) nbytes++;

    return(nbytes);
}

int write_compressed_block(int ofile,unsigned char *buf,int nbytes) {
    int errsv,nwrite;

    nwrite = write(ofile,buf,nbytes);
    errsv = errno;
    if (nwrite != nbytes) {
	fprintf(stderr,"write_compressed_block: Error: could not write to binfile. errno=%d\n",errsv);
	return(-1);
    }
    return(0);
}

/* write up to 16 bits */
void WriteBits(unsigned short inputBits, int numBits, unsigned char *obuf) {
    int ii;
    int bitsWritten = 0;

    void WriteToBuffer(unsigned int bit, unsigned char *obuf);

    if (numBits > 16) numBits = 16;
    for (ii = 0; ii < numBits; ii++) {
	unsigned int bit = (inputBits >> ii) & 0x1;
	WriteToBuffer(bit, obuf);
	bitsWritten++;
    }
}

void WriteToBuffer(unsigned int bit, unsigned char *obuf) {

    void WriteFlush(unsigned char *obuf);

    if (bitsInBuffer == 8) WriteFlush(obuf);
    if (bit != 0)   /* only write 1 bits. bits is initialized to 0 */
	bits |= (unsigned char)(bit << bitsInBuffer);
    bitsInBuffer++;
}

void WriteFlush(unsigned char *obuf) {
    if (bitsInBuffer != 0) {
	obuf[bytesWritten++] = bits;
    }
    bits = 0;
    bitsInBuffer = 0;
}

