/* 32-bit -> 16-bit integer compression */

/* Read photocurrent block size, binwidth and output-file arguments from the command-line.
 * Read a stream of floating-point photocurrents, in units of e-/sec.
 * Then, apply a binwidth scaling factor to integerize each photocurrent.
 * Then, for each block of photocurrents, calculate and apply an offset
 * such that the minimum value in each block is zero.
 * Then, pack the photocurrents in each block into 16-bit floating-point format.
 * Then, write the offsets and the packed blocks to a binary file.
 *
 * Input photocurrents are read from stdin, in text format.
 *      It does not matter how many photocurrents per line of input, but the imread program writes 2048/line.
 * Diagnostic text output is written to stdout (min scaled_input packed_input).
 * Packed photocurrents and offsets are written to a binary file specified on the command line.
 *
 * Usage: pack16b_v02 nperblock scale binfile
 */ 

/*
Packing algorithm is a modified biased exponent, hidden one algorithm
with a 12 bit mantissa and a 4 bit exponent. Numbers up to 2^12 are
unchanged, and numbers up to 2^13 unpack with no error.

Numbers up to 4095 * 2^15 (134,184,960) can be packed.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
		     
#include "packsubs.c"

#define IMSIZE 2048 * 2048

int main(argc,argv)
    int argc;
    char *argv[];
{
    float offset;
    float binwidth;
    int binfile;
    int nperblock;
    int nwrite;
    float input;
    unsigned short *output;
    int *scaled_input;
    int *minimums;		/* make the minimums signed, since we may have slightly negative photocurrents */
    int *minimumsX;
    int ii,jj;
    int idx = 0;

    unsigned int pack_rate();


    scaled_input = (int *) malloc(IMSIZE * sizeof(int));
    output = (unsigned short *) malloc(IMSIZE * sizeof(unsigned short));

    if (argc!=4) {
       fprintf(stderr,"Usage: %s nperblock scale binfile\n",argv[0]);
       exit(1);
    }

    nperblock = atoi(argv[1]);
    binwidth = (float)atof(argv[2]);

    if (nperblock<0) {
        fprintf(stderr,"Error: %s npix_per_block %d is negative\n",argv[0],nperblock);
        exit(1);
    }
    if (nperblock>IMSIZE) {
        fprintf(stderr,"Error: %s npix_per_block %d is greater than image size %d\n",argv[0],nperblock,IMSIZE);
        exit(1);
    }
    if (nperblock>65535) {
        fprintf(stderr,"Error: %s npix_per_block %d is greater than 65535\n",argv[0],nperblock);
        exit(1);
    }
    if (IMSIZE % nperblock) {
        fprintf(stderr,"Error: %s image size %d must be divisible by npix_per_block %d\n",argv[0],IMSIZE,nperblock);
        exit(1);
    }


    fprintf(stdout,"npb binwidth: %d      %.4f\n",nperblock,binwidth);  

    minimums = (int *) malloc((IMSIZE/nperblock) * sizeof(int));
    minimumsX = (int *) malloc((IMSIZE/nperblock) * sizeof(int));

    if ((binfile=open(argv[3],O_CREAT | O_RDWR | O_TRUNC,S_IRUSR | S_IWUSR)) == -1) {
       fprintf(stderr,"Error: %s could not open binfile\n",argv[0]);
       exit(1);
    }
    
    while (fscanf(stdin,"%f",&input) != EOF){
       if (input >= 0.0)
          scaled_input[idx] = (int)((input / binwidth) +0.5);
       else	  
	  scaled_input[idx] = (int)((input / binwidth) -0.5);
       if(++idx>IMSIZE) {
          fprintf(stderr,"Error: %s input greater than 2048x2048\n",argv[0]);
          exit(1);
       }
    }
    if(idx!=IMSIZE) {
       fprintf(stderr,"Error: %s input not equal to 2048x2048\n",argv[0]);
       exit(1);
    }

    idx=0;
    for (ii=0;ii<IMSIZE/nperblock;ii++) {
	int min=0xFFFFF;

	for (jj=0;jj<nperblock;jj++) {
	    if (scaled_input[idx+jj] < min) {
	       min=scaled_input[idx+jj];
	    }
	}
	minimumsX[ii]=min;
	/* We want each min in the final compressed data to fit into 12 bits, including a sign bit */	
	if(min>2047) {
/*	    fprintf(stderr,"Warning: %s minimum for block %d is %d. Setting it to 2047\n",argv[0],ii,min); */
	    min=2047;
	}
	if(min<-2047) {
/*	    fprintf(stderr,"Warning: %s minimum for block %d is %d. Setting it to -2047\n",argv[0],ii,min); */
	    min=-2047;
	}
	minimums[ii]=min;

	for (jj=0;jj<nperblock;jj++) {
	    int tmp = scaled_input[idx+jj] - min;
	    output[idx+jj] = (unsigned short)pack_rate(tmp);
/*	    fprintf(stdout,"%d      %d   %d  %d\n",ii,min,scaled_input[idx+jj],output[idx+jj]);  */
	}

	idx += nperblock;
	fprintf(stdout,"%d      %d    %5d %5d",ii,idx,minimums[ii],minimumsX[ii]);
	if (minimums[ii] !=minimumsX[ii]) {
	   fprintf(stdout,"  XXX\n");
	} else {
	   fprintf(stdout,"\n");
	}
    }
    if(idx!=IMSIZE) {
       fprintf(stderr,"Error: %s output not equal to 2048x2048\n",argv[0]);
       exit(1);
    }

    nwrite = write(binfile,minimums,(IMSIZE/nperblock) * sizeof(int));
    int errsv = errno;
    if (nwrite !=((IMSIZE/nperblock) * sizeof(int))) {
       fprintf(stderr,"Error: %s could not write mins to binfile. errno=%d\n",argv[0],errsv);
       exit(1);
    }

    nwrite = write(binfile,output,idx*sizeof(unsigned short));
    errsv = errno;
    if (nwrite !=(idx*sizeof(unsigned short))) {
       fprintf(stderr,"Error: %s could not write data to binfile. errno=%d\n",argv[0],errsv);
       exit(1);
    }

    close(binfile);
}

